@@ -4,7 +4,7 @@ import subprocess
 
 import MDAnalysis as mda
 import numpy as np
-from shapeGMMTorch import torch_align
+from shapeGMMTorch import align
 import torch
 
 import mdance.tools.esim as esim
@@ -41,6 +41,8 @@ def mean_sq_dev(matrix, N_atoms):
     32.8
     """
     N = len(matrix)
+    if N == 1:
+        return 0
     sq_data = matrix ** 2
     c_sum = np.sum(matrix, axis=0)
     sq_sum = np.sum(sq_data, axis=0)
@@ -84,6 +86,8 @@ def msd_condensed(c_sum, sq_sum, N, N_atoms):
     >>> bts.msd_condensed(c_sum, sq_sum, N=5, N_atoms=1)
     32.8
     """
+    if N == 1:
+        return 0
     msd = np.sum(2 * (N * sq_sum - c_sum ** 2)) / (N ** 2)
     norm_msd = msd / N_atoms
     return norm_msd
@@ -153,6 +157,8 @@ def extended_comparison(matrix, data_type='full', metric='MSD', N=None,
     if data_type == 'full':
         if not isinstance(matrix, np.ndarray):
             raise TypeError('data must be a numpy.ndarray')
+        if matrix.ndim != 2:
+            raise ValueError('Input must be numpy ndarray of shape (n_samples, n_features)')
         c_sum = np.sum(matrix, axis=0)
         if metric == 'MSD':
             sq_data = matrix ** 2
@@ -259,7 +265,7 @@ def calculate_medoid(matrix, metric, N_atoms=1):
     >>> bts.calculate_medoid(X, metric='MSD', N_atoms=1)
     2
     """
-    return np.argmax(calculate_comp_sim(matrix, metric, N_atoms=1))
+    return np.argmax(calculate_comp_sim(matrix, metric, N_atoms))
 
 
 def calculate_outlier(matrix, metric, N_atoms=1):
@@ -289,7 +295,7 @@ def calculate_outlier(matrix, metric, N_atoms=1):
     >>> bts.calculate_outlier(X, metric='MSD', N_atoms=1)
     4
     """
-    return np.argmin(calculate_comp_sim(matrix, metric, N_atoms=1))
+    return np.argmin(calculate_comp_sim(matrix, metric, N_atoms))
 
 
 def trim_outliers(matrix, n_trimmed, metric, N_atoms, criterion='comp_sim'):
@@ -368,7 +374,7 @@ def trim_outliers(matrix, n_trimmed, metric, N_atoms, criterion='comp_sim'):
 
 
 def diversity_selection(matrix, percentage: int, metric, N_atoms=1, 
-                        start='medoid'):
+                        method='strat', start='medoid'):
     """*O(N)* method of selecting the most diverse subset of a data 
     matrix using the complementary similarity. 
     
@@ -377,13 +383,19 @@ def diversity_selection(matrix, percentage: int, metric, N_atoms=1,
     matrix : array-like of shape (n_samples, n_features)
         A feature array.
     percentage : int
-        Percentage of the data to select.
+        If ``method='strat'``, percentage indicates how many bins of stratified 
+        data will be generated. If ``method='comp_sim'``, percentage indicates the
+        percentage of data to be selected.
     metric : str, default='MSD'
         The metric to when calculating distance between *n* objects in an array. 
         It must be an options allowed by :func:`mdance.tools.bts.extended_comparison`.
     N_atoms : int, default=1
         Number of atoms in the system used for normalization.
         ``N_atoms=1`` for non-Molecular Dynamics datasets.
+    method : {'strat', 'comp_sim'}, default='strat'
+        The method to use for diversity selection. ``strat``: stratified
+        sampling. ``comp_sim``: maximizing the MSD between the selected
+        objects and the rest of the data.
     start : {'medoid', 'outlier', 'random', list}, default='medoid'
         The initial seed for initiating diversity selection. Either 
         from one of the options or a list of indices are valid inputs.
@@ -404,55 +416,73 @@ def diversity_selection(matrix, percentage: int, metric, N_atoms=1,
     --------
     >>> from mdance.tools import bts
     >>> import numpy as np
-    >>> X = np.array([[1, 2], [2, 2], [2, 3], [8, 7], [8, 8]])
-    >>> bts.diversity_selection(X, percentage=10, metric='MSD', N_atoms=1)
-    [2]
+    >>> X = np.array([[1, 2], [2, 2], [2, 3], [8, 7], [8, 8], [2, 9], [1, 8], [2, 7]])
+    >>> bts.diversity_selection(X, percentage=30, metric='MSD', N_atoms=1)
+    [7 4]
     """
     n_total = len(matrix)
-    total_indices = np.array(range(n_total))
-    
-    if start =='medoid':
-        seed = calculate_medoid(matrix, metric=metric, N_atoms=N_atoms)
-        selected_n = [seed]
-    elif start == 'outlier':
-        seed = calculate_outlier(matrix, metric=metric, N_atoms=N_atoms)
-        selected_n = [seed]
-    elif start == 'random':
-        seed = random.randint(0, n_total - 1)
-        selected_n = [seed]
-    elif isinstance(start, list):
-        selected_n = start
-    else:
-        raise ValueError('Select a correct starting point: medoid, outlier, \
-                         random or outlier')
-
-    n = len(selected_n)
     n_max = int(np.floor(n_total * percentage / 100))
     if n_max > n_total:
-        raise ValueError('Percentage is too high')
-    selection = [matrix[i] for i in selected_n] 
-    selection = np.array(selection)
-    selected_condensed = np.sum(selection, axis=0)
-    if metric == 'MSD':
-        sq_selection = selection ** 2
-        sq_selected_condensed = np.sum(sq_selection, axis=0)
+        raise ValueError('Percentage is too high for the given matrix size')
     
-    while len(selected_n) < n_max:
-        select_from_n = np.delete(total_indices, selected_n)
-        if metric == 'MSD':
-            new_index_n = get_new_index_n(matrix, metric=metric, 
-                                          selected_condensed=selected_condensed,
-                                          sq_selected_condensed=sq_selected_condensed, 
-                                          n=n, select_from_n=select_from_n, 
-                                          N_atoms=N_atoms)
-            sq_selected_condensed += matrix[new_index_n] ** 2
+    if method == 'strat':
+        if n_max <= 1:
+            raise ValueError('Percentage is too low for the given matrix size')
+        if n_max == 1:
+            indices_to_select = [0]
         else:
-            new_index_n = get_new_index_n(matrix, metric=metric, 
-                                          selected_condensed=selected_condensed, 
-                                          n=n, select_from_n=select_from_n)
-        selected_condensed += matrix[new_index_n]
-        selected_n.append(new_index_n)
+            step = (n_total - 1) / (n_max - 1)
+            indices_to_select = np.round(np.arange(n_max) * step).astype(int)
+            indices_to_select[0] = 0
+        comp_sims = calculate_comp_sim(matrix, metric=metric, N_atoms=N_atoms)
+        sorted_comps = np.argsort(-comp_sims, kind='stable')
+        selected_n = sorted_comps[indices_to_select].tolist()
+        
+    elif method == 'comp_sim':
+        total_indices = np.array(range(n_total))
+        
+        if start =='medoid':
+            seed = calculate_medoid(matrix, metric=metric, N_atoms=N_atoms)
+            selected_n = [seed]
+        elif start == 'outlier':
+            seed = calculate_outlier(matrix, metric=metric, N_atoms=N_atoms)
+            selected_n = [seed]
+        elif start == 'random':
+            seed = random.randint(0, n_total - 1)
+            selected_n = [seed]
+        elif isinstance(start, list):
+            selected_n = start
+        else:
+            raise ValueError('Select a correct starting point: medoid, outlier, \
+                            random or outlier')
+
         n = len(selected_n)
+        
+        selection = [matrix[i] for i in selected_n] 
+        selection = np.array(selection)
+        selected_condensed = np.sum(selection, axis=0)
+        if metric == 'MSD':
+            sq_selection = selection ** 2
+            sq_selected_condensed = np.sum(sq_selection, axis=0)
+        
+        while len(selected_n) < n_max:
+            select_from_n = np.delete(total_indices, selected_n)
+            if metric == 'MSD':
+                new_index_n = get_new_index_n(matrix, metric=metric, 
+                                            selected_condensed=selected_condensed,
+                                            sq_selected_condensed=sq_selected_condensed, 
+                                            n=n, select_from_n=select_from_n, 
+                                            N_atoms=N_atoms)
+                sq_selected_condensed += matrix[new_index_n] ** 2
+            else:
+                new_index_n = get_new_index_n(matrix, metric=metric, 
+                                            selected_condensed=selected_condensed, 
+                                            n=n, select_from_n=select_from_n)
+            selected_condensed += matrix[new_index_n]
+            selected_n.append(new_index_n)
+            n = len(selected_n)
+    else:
+        raise ValueError('Select a correct sampling method: strat or comp_sim')
     return selected_n
 
 
@@ -551,16 +581,14 @@ def align_traj(data, N_atoms, align_method=None):
     device = torch.device('cpu')
     dtype = torch.float32
     traj_tensor = torch.tensor(data, device=device, dtype=dtype)
-    torch_align.torch_remove_center_of_geometry(traj_tensor)
+    traj_tensor = align.remove_center_of_geometry(traj_tensor)
     
     if align_method == 'uni' or align_method == 'uniform':
-        uniform_aligned_traj_tensor, uniform_avg_tensor, uniform_var_tensor = torch_align.torch_iterative_align_uniform(
-            traj_tensor, device=device, dtype=dtype, verbose=True)
+        uniform_aligned_traj_tensor, uniform_avg_tensor, uniform_var_tensor = align.maximum_likelihood_uniform_alignment(traj_tensor, verbose=True)
         aligned_traj_numpy = uniform_aligned_traj_tensor.cpu().numpy()
     
     elif align_method == 'kron' or align_method == 'kronecker':
-        kronecker_aligned_traj_tensor, kronecker_avg_tensor, kronecker_precision_tensor, kronecker_lpdet_tensor = torch_align.torch_iterative_align_kronecker(
-            traj_tensor, device=device, dtype=dtype, verbose=True)
+        kronecker_aligned_traj_tensor, kronecker_avg_tensor, kronecker_precision_tensor, kronecker_lpdet_tensor = align.maximum_likelihood_kronecker_alignment(traj_tensor, verbose=True)
         aligned_traj_numpy = kronecker_aligned_traj_tensor.cpu().numpy()
     
     else:
@@ -571,7 +599,7 @@ def align_traj(data, N_atoms, align_method=None):
 
 
 def equil_align(indices, sieve, input_top, input_traj, mdana_atomsel, cpptraj_atomsel, ref_index):
-    """ Aligns the frames in the trajectory to the reference frame.
+    """Aligns the frames in the trajectory to the reference frame.
     
     Parameters
     ----------
@@ -616,4 +644,98 @@ def equil_align(indices, sieve, input_top, input_traj, mdana_atomsel, cpptraj_at
     os.remove('cpptraj.in')
     os.remove('unaligned_traj.pdb')
     os.remove('aligned_traj.pdb')
-    return aligned_traj_numpy
\ No newline at end of file
+    return aligned_traj_numpy
+
+
+def rep_sample(data, metric='MSD', N_atoms=1, n_bins=10, n_samples=100, 
+               hard_cap=True):
+    """Representative sampling according to comp_sim values.
+    
+    Divides the range of comp_sim values in nbins and then
+    uniformly selects n_samples molecules, consecutively
+    taking one from each bin
+    
+    Parameters
+    ----------
+    data : array-like of shape (n_samples, n_features)
+        The data to be sampled.
+    metric : str, default='MSD'
+        The metric to be used for the comparison.
+    N_atoms : int, default=1
+        Number of atoms in the Molecular Dynamics (MD) system. ``N_atom=1`` 
+        for non-MD systems.
+    n_bins : int, default=10
+        Number of bins to divide the comp_sim values.
+    n_samples : int, default=100
+        Number of samples to be selected.
+    hard_cap : bool, default=True
+        If True, the number of samples will be exactly n_samples.
+        If False, the number of samples may not be exactly n_samples.
+        
+    Returns
+    -------
+    sampled_mols : list
+        List of indices of the sampled objects in the original data
+    """
+    n = len(data)
+    if n_samples < 1:
+        n_samples = int(n * n_samples)
+    cs = calculate_comp_sim(data, metric=metric, N_atoms=N_atoms)
+    tups = []
+    for i, comp in enumerate(cs):
+        tups.append((i, comp))
+    comp_sims = np.sort(cs)
+    mi = np.min(comp_sims)
+    ma = np.max(comp_sims)
+    D = ma - mi
+    step = D / n_bins
+    bins = []
+    indices = np.array(list(range(n)))
+    for i in range(n_bins - 1):
+        low = mi + i * step
+        up = mi + (i + 1) * step
+        bins.append(indices[(comp_sims >= low) * (comp_sims < up)])
+    bins.append(indices[(comp_sims >= up) * (comp_sims <= ma)])
+    order_sampled = []
+    i = 0
+    while len(order_sampled) < n_samples:
+        for b in bins:
+            if len(b) > i:
+                order_sampled.append(b[i])
+                if hard_cap:
+                    if len(order_sampled) >= n_samples:
+                        break
+            else:
+                pass
+        i += 1
+    tups.sort(key = lambda tups : tups[1])
+    sampled_mols = []
+    for i in order_sampled:
+        sampled_mols.append(tups[i][0])
+    return sampled_mols
+
+
+def refine_dis_matrix(matrix):
+    """Refine a distance matrix by setting the diagonal to zero and
+    symmetrizing the matrix.
+    
+    Parameters
+    ----------
+    matrix : array-like of shape (n_samples, n_features)
+        A feature array.
+    
+    Returns
+    -------
+    numpy.ndarray
+        A refined 2D matrix.
+    """
+    distances = np.array(matrix)
+    if distances.ndim != 2:
+        raise ValueError('Matrix must be 2D')
+    if distances.shape[0] != distances.shape[1]:
+        raise ValueError('Matrix must be square')
+    distances += distances.T
+    distances *= 0.5
+    distances -= np.min(distances)
+    np.fill_diagonal(distances, 0)
+    return distances
\ No newline at end of file
